import math
import sys

from PIL import Image, ImageDraw


class TrafficSignGenerator:
    def __init__(self, size=400):
        self.size = size
        self.border_width = size // 8
        self.line_width = size // 12


    def _create_base_circle(self):
        """Create a basic white circle with red border"""
        # Create a new image with white background
        image = Image.new('RGB', (self.size, self.size), 'white')
        draw = ImageDraw.Draw(image)

        # Draw red circle border
        draw.ellipse([0, 0, self.size-1, self.size-1], outline='red', width=self.border_width)

        return image, draw


    def _paste_symbol(self, image, symbol_path, region):
        """Paste a symbol into specified region of the circle"""
        if not symbol_path:
            return

        # Load and resize symbol
        symbol = Image.open(symbol_path).convert('RGBA')

        # Calculate region dimensions
        x1, y1, x2, y2 = region
        region_width = x2 - x1
        region_height = y2 - y1

        # Resize symbol to fit region (maintaining aspect ratio)
        symbol_ratio = min(region_width / symbol.width, region_height / symbol.height)
        new_size = (int(symbol.width * symbol_ratio * 0.7),
                   int(symbol.height * symbol_ratio * 0.7))
        symbol = symbol.resize(new_size, Image.Resampling.LANCZOS)

        # Calculate center position
        paste_x = x1 + (region_width - symbol.width) // 2
        paste_y = y1 + (region_height - symbol.height) // 2

        # Paste symbol
        image.paste(symbol, (paste_x, paste_y), symbol)


    def create_full_circle(self, symbol_path=None, output_path="out.png"):
        """Create a full circle sign with optional symbol"""
        image, _ = self._create_base_circle()

        if symbol_path:
            # Define region for symbol (leaving margin for border)
            margin = 0 # self.border_width * 2
            region = [margin, margin, self.size, self.size]
            self._paste_symbol(image, symbol_path, region)

        image.save(output_path)


    def create_split_circle(self, top_symbol=None, bottom_symbol=None, output_path="out.png"):
        """Create a circle split horizontally with two optional symbols"""
        image, draw = self._create_base_circle()

        # Draw horizontal line
        draw.line([(0, self.size//2), (self.size, self.size//2)],
                 fill='red', width=self.line_width)

        margin = self.size // 10
        # Define regions for symbols
        top_region = [0, margin,
                     self.size, self.size // 2 ]
        bottom_region = [0, self.size // 2 , self.size, self.size - margin]

        self._paste_symbol(image, top_symbol, top_region)
        self._paste_symbol(image, bottom_symbol, bottom_region)

        image.save(output_path)


    def create_third_circle(self, symbols=None, output_path="out.png"):
        """Create a circle split into thirds with optional symbols"""
        image, draw = self._create_base_circle()

        # Calculate points for the Y-shaped division
        center_x = self.size // 2
        center_y = self.size // 2
        radius = self.size // 2

        # Draw Y-shaped division
        for angle in [60, 180, 300]:
            rad = math.radians(angle)
            end_x = center_x + radius * math.cos(rad)
            end_y = center_y + radius * math.sin(rad)
            draw.line([(center_x, center_y), (end_x, end_y)],
                     fill='red', width=self.line_width)

        margin = self.size // 10
        regions = [
            # Top left region
            [margin * 2, margin, self.size//2 + margin//2, self.size//2],
            # Bottom left region
            [margin * 2, self.size//2, self.size//2 + margin//2, self.size-margin],
            # Right region
            [self.size//2 + margin, self.size//4, self.size-margin, self.size*3//4]
        ]

        if symbols:
            for i, symbol_path in enumerate(symbols[:3]):
                if symbol_path:
                    self._paste_symbol(image, symbol_path, regions[i])

        image.save(output_path)


if __name__ == "__main__":
    generator = TrafficSignGenerator(size=1200)

    match len(sys.argv):
        case 0 | 1:
            print("Usage:")
            print(f"python3 {sys.argv[0]} NONE")
            print(f"python3 {sys.argv[0]} [png_path]")
            print(f"python3 {sys.argv[0]} [png_path] [png_path2]")
            print(f"python3 {sys.argv[0]} [png_path] [png_path2] [png_path3]")
        case 2:
            if sys.argv[1] == 'NONE':
                generator.create_full_circle(None, "out.png")
            else:
                generator.create_full_circle(sys.argv[1], "out.png")
        case 3:
            generator.create_split_circle(sys.argv[1], sys.argv[2], "out.png")
        case 4:
            generator.create_third_circle([sys.argv[1], sys.argv[2], sys.argv[3]], "out.png")
