import json
from typing import Callable, Optional, Set, Tuple

# Malý šachový slovníček cs-en:
# Figurka = Piece
# Věž = Rook
# Jezdec = Knight
# Král = King
# Střelec = Bishop
# Tah = Move
# Šachovnice = Chess Board
# Souřadnice = Coordinates
# Mřížka = Grid

Position = Tuple[int, int]

class Piece:
    @staticmethod
    def rook_moves(position: Position) -> Set[Position]:
        """Get all possible single moves for a rook from given position"""
        row, col = position
        moves = set()

        # Horizontal moves
        for new_col in range(8):
            if new_col != col:
                moves.add((row, new_col))

        # Vertical moves
        for new_row in range(8):
            if new_row != row:
                moves.add((new_row, col))

        return moves


    @staticmethod
    def bishop_moves(position: Position) -> Set[Position]:
        """Get all possible single moves for a bishop from given position"""
        row, col = position
        moves = set()

        # All diagonal directions
        for i in range(-7, 8):
            if i != 0:
                # Top-left to bottom-right diagonal
                new_pos1 = (row + i, col + i)
                if is_valid_position(new_pos1):
                    moves.add(new_pos1)

                # Top-right to bottom-left diagonal
                new_pos2 = (row + i, col - i)
                if is_valid_position(new_pos2):
                    moves.add(new_pos2)
        return moves


    @staticmethod
    def knight_moves(position: Position) -> Set[Position]:
        """Get all possible single moves for a knight from given position"""
        row, col = position
        possible_moves = [
            (-2, -1), (-2, 1), (-1, -2), (-1, 2),
            (1, -2), (1, 2), (2, -1), (2, 1)
        ]

        moves = set()
        for move_row, move_col in possible_moves:
            new_pos = (row + move_row, col + move_col)
            if is_valid_position(new_pos):
                moves.add(new_pos)

        return moves


    @staticmethod
    def king_moves(position: Position) -> Set[Position]:
        """Get all possible single moves for a king from given position"""
        row, col = position
        possible_moves = [
            (-1, -1), (-1, 0), (-1, 1),
            (0, -1),          (0, 1),
            (1, -1),  (1, 0),  (1, 1)
        ]

        moves = set()
        for move_row, move_col in possible_moves:
            new_pos = (row + move_row, col + move_col)
            if is_valid_position(new_pos):
                moves.add(new_pos)

        return moves


def is_valid_position(position: Position) -> bool:
    """Check if the position is within the chessboard bounds"""
    row, col = position
    return 0 <= row < 8 and 0 <= col < 8


def convert_position(position: str) -> Position:
    """Convert chess position (e.g. 'a1') to grid coordinates (0,0)"""
    col = ord(position[0].lower()) - ord('a')
    row = int(position[1]) - 1
    return (row, col)


def convert_to_chess_notation(position: Position) -> str:
    """Convert grid coordinates to chess notation"""
    row, col = position
    return f"{chr(col + ord('a'))}{row + 1}"


def bfs(start: Position, end: Position, piece_sequence: list[Callable]) -> Optional[list[Position]]:
    """Search for a shortest path between start and end"""
    turn = 0
    to_visit_next = [start]
    previous_position = {}
    visited = set([start])  # Track visited positions to avoid loops

    while to_visit_next:
        sequence_position = turn % len(piece_sequence)
        to_visit = to_visit_next.copy()
        to_visit_next.clear()

        for position in to_visit:
            # Calling the appropriate method
            method_for_this_turn = piece_sequence[sequence_position]
            can_visit = method_for_this_turn(position)

            for candidate in can_visit:
                if candidate not in visited:  # Process only unvisited positions
                    visited.add(candidate)
                    previous_position[f"{candidate}_{turn}"] = position
                    to_visit_next.append(candidate)

        if end in to_visit_next:
            # Backtrack to reconstruct the path
            pos = end
            path = [end]
            for backtrack_turn in range(turn, 0, -1):
                pos = previous_position[f"{pos}_{backtrack_turn}"]
                path.append(pos)
            path.append(start)
            path.reverse()
            return path

        turn += 1

    return None  # No path exists


def move_method(piece_type):
    """Search for a shortest path between start and end"""
    match piece_type:
        case 'vez':
            return Piece.rook_moves
        case 'kral':
            return Piece.king_moves
        case 'jezdec':
            return Piece.knight_moves
        case 'strelec':
            return Piece.bishop_moves
        case _:
            raise ValueError('Neznama figurka', piece_type)


def main():
    json_input_filename = input("Zadej cestu ke vstupnímu json souboru.\n")

    with open(json_input_filename, 'r', encoding="utf-8") as file:
        input_data = json.load(file)


    move_methods = [move_method(jmeno_figurky) for jmeno_figurky in input_data["rozpis"]]
    start = input_data["start"]
    cil = input_data["cil"]

    path = bfs(convert_position(start), convert_position(cil), move_methods)

    if path is not None:
        print(f"Nalezena nejkratší cesta délky {len(path) - 1}")
        print(",".join([convert_to_chess_notation(position) for position in path]))
    else:
        print("Žádná cesta neexistuje.")


if __name__ == "__main__":
    main()
