use std::convert::identity;
use clap::builder::ValueParser;
use clap::Parser;
use itertools::Itertools;
use rayon::prelude::*;
use std::error::Error;
use std::fmt::{Debug, Display, Formatter};
use std::iter::repeat;
use std::net::{SocketAddr, ToSocketAddrs};
use std::process::exit;
use smallvec::{smallvec, SmallVec};
use tokio::io::{AsyncBufReadExt, AsyncWriteExt, BufReader};
use tokio::net::TcpStream;
use tokio::time::Instant;

#[derive(Debug)]
enum ServerParseError {
    IO(std::io::Error),
    Empty,
}

impl From<std::io::Error> for ServerParseError {
    fn from(value: std::io::Error) -> Self {
        Self::IO(value)
    }
}

fn parse_server(s: &str) -> Result<SocketAddr, ServerParseError> {
    match s.to_socket_addrs().map(|mut iter| match iter.next() {
        None => Err(ServerParseError::Empty),
        Some(value) => Ok(value),
    }) {
        Ok(Ok(value)) => Ok(value),
        Ok(Err(error)) => Err(error),
        Err(error) => Err(error.into()),
    }
}

impl Display for ServerParseError {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::IO(err) => {
                write!(f, "Failed resolving server name: {err}")
            }
            Self::Empty => {
                write!(f, "No IP address found")
            }
        }
    }
}

impl Error for ServerParseError {}

#[derive(Debug, Parser, Clone)]
#[command(version, about, long_about = None)]
struct Args {
    /// Server address and port
    #[arg(short, long, default_value = "[::1]:1234", value_parser = ValueParser::new(parse_server))]
    server: SocketAddr,
}

trait PrintError {
    type Value;
    fn print(self) -> Self::Value;
}

impl<V, E> PrintError for Result<V, E>
where
    E: Debug,
{
    type Value = V;
    fn print(self) -> V {
        self.map_err(|err| {
            eprintln!("{err:?}");
            exit(1);
        })
        .expect("Impossible")
    }
}

impl<V> PrintError for Option<V> {
    type Value = V;
    fn print(self) -> V {
        match self {
            None => {
                eprintln!("No value in expression");
                exit(1);
            }
            Some(value) => value,
        }
    }
}

#[derive(Debug, Default, Copy, Clone, Hash, Ord, PartialOrd, Eq, PartialEq)]
struct IndexType {
    inner: u8,
}

impl IndexType {
    fn to_usize(self) -> usize {
        self.inner.into()
    }
    
    fn from_usize(value: usize) -> Self {
        Self {
            inner: value.try_into().unwrap()
        }
    }
}

impl From<usize> for IndexType {
    fn from(value: usize) -> Self {
        IndexType::from_usize(value)
    }
}

impl From<IndexType> for usize {
    fn from(value: IndexType) -> Self {
        value.to_usize()
    }
}

type GuessType = SmallVec<[IndexType; 6]>;

fn generate_all(len: usize, colors: &Vec<&str>) -> Vec<GuessType> {
    let mut states = colors
        .into_iter()
        .enumerate()
        .map(|(idx, _)| smallvec![idx.into()])
        .collect_vec();
    for _ in 1..len {
        states = states
            .iter()
            .map(|state| {
                colors.into_iter().enumerate().map(|(idx, _)| {
                    let mut new_state = state.clone();
                    new_state.push(idx.into());
                    new_state
                })
            })
            .flatten()
            .collect();
    }
    states
}

fn common(lhs: &GuessType, rhs: &GuessType) -> (usize, usize) {
    let correct_positions = lhs
        .iter()
        .zip(rhs.iter())
        .filter(|(lhs, rhs)| lhs == rhs)
        .count();
    let incorrect_positions = lhs.into_iter()
        .scan(vec![], |done_colors, &color| {
            if !done_colors.contains(&color) {
                done_colors.push(color);
                let lhs_count = lhs.into_iter()
                    .filter(|&&x| x == color)
                    .count();
                let rhs_count = rhs.into_iter()
                    .filter(|&&x| x == color)
                    .count();
                Some(Some(lhs_count.min(rhs_count)))
            } else {
                Some(None)
            }
        })
        .filter_map(identity)
        .sum::<usize>() - correct_positions;
    (correct_positions, incorrect_positions)
}

fn is_possible(
    reference: &GuessType,
    other: &GuessType,
    correct_positions: usize,
    incorrect_positions: usize,
) -> bool {
    let (same, swapped) = common(reference, other);
    same == correct_positions && swapped == incorrect_positions
}

fn best_guess<'a>(
    states: &Vec<GuessType>,
    remaining_states: &Vec<GuessType>,
    len: usize,
) -> GuessType {
    states
        .par_iter()
        .min_by_key(|&state| {
            (0..len)
                .cartesian_product(0..len)
                .filter(|(a, b)| *a + *b <= len)
                .map(|(correct, incorrect)| {
                    remaining_states
                        .into_iter()
                        .filter(|&other| is_possible(state, other, correct, incorrect))
                        .count()
                })
                .max()
        })
        .expect("Impossible")
        .clone()
}

fn translate_names(state: &GuessType, colors: &Vec<&str>) -> String {
    state.into_iter().map(|x| colors[x.to_usize()]).join(",")
}

#[tokio::main]
async fn main() {
    let args = Args::parse();
    let mut stream = TcpStream::connect(args.server).await.print();
    let (read, mut write) = stream.split();
    let mut read = BufReader::new(read);
    let mut buffer = String::new();
    read.read_line(&mut buffer).await.print();
    if buffer != "Welcome\n" {
        eprintln!("Initial message was incorrect: {buffer:?}");
        exit(1);
    }
    buffer.clear();
    read.read_line(&mut buffer).await.print();
    let len: usize = buffer.trim().parse().print();
    buffer.clear();
    let mut colors_buffer = String::new();
    read.read_line(&mut colors_buffer).await.print();
    let colors = colors_buffer.split(',').map(str::trim).collect_vec();
    println!("Sequence contains {len} colors out of {colors:?}");
    let mut remaining_states = generate_all(len, &colors);
    let mut attempts = 0;
    while !remaining_states.is_empty() {
        attempts += 1;
        println!("Number of possible states is {}", remaining_states.len());
        let start = Instant::now();
        let guess = if attempts == 1 {
            repeat((0..colors.len()).map(|x| [x, x].into_iter()))
                .flatten()
                .flatten()
                .take(len)
                .map(Into::into)
                .collect()
        } else {
            best_guess(&remaining_states, &remaining_states, len)
        };
        println!(
            "Best guess took {}s",
            (Instant::now() - start).as_secs_f32()
        );
        let guess_string = translate_names(&guess, &colors);
        println!("Trying `{}`", guess_string);
        write.write_all(guess_string.as_bytes()).await.print();
        write.write_all("\n".as_bytes()).await.print();
        buffer.clear();
        read.read_line(&mut buffer).await.print();
        if buffer == "Correct\n" {
            println!(
                "The guess `{}` was correct after {attempts} attempts!",
                guess_string
            );
            exit(0);
        }
        let (lhs, rhs) = buffer.trim().split_once(',').print();
        let correct_positions: usize = lhs.parse().print();
        let incorrect_positions: usize = rhs.parse().print();
        println!("Guess contains {correct_positions} correct positions and {incorrect_positions} incorrect positions");
        remaining_states
            .retain(|state| is_possible(&guess, state, correct_positions, incorrect_positions));
    }
    eprintln!("Could not find solution");
    exit(1);
}

#[cfg(test)]
mod tests {
    use super::*;

    fn check_equivalency_common(lhs: &GuessType, rhs: &GuessType, result: (usize, usize)) {
        assert_eq!(common(lhs, rhs), result);
        assert_eq!(common(rhs, lhs), result);
    }

    fn to_indexes(idxs: Vec<usize>) -> GuessType {
        idxs.into_iter()
            .map(Into::into)
            .collect()
    }

    #[test]
    fn test_common() {
        check_equivalency_common(&to_indexes(vec![0, 1, 2, 3]), &to_indexes(vec![0, 1, 2, 3]), (4, 0));
        check_equivalency_common(&to_indexes(vec![0, 1, 2, 3]), &to_indexes(vec![0, 1, 2, 2]), (3, 0));
        check_equivalency_common(&to_indexes(vec![0, 1, 2, 3]), &to_indexes(vec![0, 2, 1, 3]), (2, 2));
        check_equivalency_common(&to_indexes(vec![0, 1, 2, 3]), &to_indexes(vec![0, 2, 1, 2]), (1, 2));
        check_equivalency_common(&to_indexes(vec![0, 1, 2, 3]), &to_indexes(vec![0, 1, 1, 2]), (2, 1));
        check_equivalency_common(&to_indexes(vec![0, 1, 2, 3]), &to_indexes(vec![0, 0, 0, 0]), (1, 0));
    }

    #[test]
    /// Discovered a bug with the following guesses
    /// yellow,black,green,black -> 2,0
    /// yellow,white,green,black -> 1,2
    /// yellow,black,white,red -> 2,2
    /// yellow,black,black,black => This should not have been considered as an option
    fn bug_001() {
        assert!(!is_possible(&to_indexes(vec![3, 4, 6, 0]), &to_indexes(vec![3, 4, 4, 4]), 2, 2));
        let mut options =
            generate_all(4, &vec!["red", "green", "blue", "yellow", "black", "white"]);
        options.retain(|state| is_possible(&to_indexes(vec![3, 4, 1, 4]), state, 2, 0));
        options.retain(|state| is_possible(&to_indexes(vec![3, 5, 1, 4]), state, 1, 2));
        options.retain(|state| is_possible(&to_indexes(vec![3, 4, 5, 0]), state, 2, 2));
        assert!(!options.contains(&to_indexes(vec![3, 4, 4, 4])));
        assert!(options.contains(&to_indexes(vec![3, 4, 0, 5])));
        assert_eq!(options.len(), 1);
    }
}
